import 'dart:async';
import 'package:bloc/bloc.dart';
import 'package:equatable/equatable.dart';
import 'package:syncrow_web/pages/device_managment/all_devices/models/device_status.dart';
import 'package:syncrow_web/pages/device_managment/water_heater/models/water_heater_status_model.dart';
import 'package:syncrow_web/services/devices_mang_api.dart';

part 'water_heater_event.dart';
part 'water_heater_state.dart';

class WaterHeaterBloc extends Bloc<WaterHeaterEvent, WaterHeaterState> {
  WaterHeaterBloc() : super(WaterHeaterInitial()) {
    on<WaterHeaterFetchStatusEvent>(_fetchWaterHeaterStatus);
    on<ToggleWaterHeaterEvent>(_controlWaterHeater);
    on<UpdateScheduleEvent>(_updateScheduleEvent);
    on<StopScheduleEvent>(_stopScheduleEvent);
  }

  late WaterHeaterStatusModel deviceStatus;
  Timer? _timer;

  FutureOr<void> _updateScheduleEvent(
    UpdateScheduleEvent event,
    Emitter<WaterHeaterState> emit,
  ) async {
    final currentState = state as WaterHeaterScheduleViewState;

    final countdownRemaining = currentState.isActive
        ? currentState.countdownRemaining
        : Duration(hours: event.hours, minutes: event.minutes);

    emit(WaterHeaterScheduleViewState(
      scheduleMode: event.scheduleMode,
      hours: countdownRemaining!.inHours,
      minutes: countdownRemaining.inMinutes % 60,
      isActive: currentState.isActive,
      countdownRemaining: countdownRemaining,
    ));

    if (currentState.isActive) {
      _startCountdown(countdownRemaining, emit);
    }
  }

  FutureOr<void> _controlWaterHeater(
      ToggleWaterHeaterEvent event, Emitter<WaterHeaterState> emit) async {
    final oldValue = _getValueByCode(event.code);

    _updateLocalValue(event.code, event.value, emit);

    emit(WaterHeaterDeviceStatusLoaded(deviceStatus));

    final success = await _runDebounce(
      deviceId: event.deviceId,
      code: event.code,
      value: event.value,
      oldValue: oldValue,
      emit: emit,
    );

    if (success && (event.code == "countdown_1" || event.code == "switch_inching")) {
      final countdownDuration = Duration(seconds: event.value);

      emit(WaterHeaterScheduleViewState(
        scheduleMode: deviceStatus.scheduleMode,
        hours: countdownDuration.inHours,
        minutes: (countdownDuration.inMinutes % 60),
        isActive: true,
        countdownRemaining: countdownDuration,
      ));

      _startCountdown(countdownDuration, emit);
    }
  }

  Future<bool> _runDebounce({
    required String deviceId,
    required String code,
    required dynamic value,
    required dynamic oldValue,
    required Emitter<WaterHeaterState> emit,
  }) async {
    final completer = Completer<bool>();

    if (_timer != null) {
      _timer!.cancel();
    }

    _timer = Timer(const Duration(milliseconds: 500), () async {
      try {
        final status = await DevicesManagementApi().deviceControl(
          deviceId,
          Status(code: code, value: value),
        );

        if (!status) {
          _revertValueAndEmit(deviceId, code, oldValue, emit);
          completer.complete(false);
        } else {
          completer.complete(true);
        }
      } catch (e) {
        _revertValueAndEmit(deviceId, code, oldValue, emit);
        completer.complete(false);
      }
    });

    return completer.future;
  }

  void _revertValueAndEmit(
      String deviceId, String code, dynamic oldValue, Emitter<WaterHeaterState> emit) {
    _updateLocalValue(code, oldValue, emit);
    emit(WaterHeaterDeviceStatusLoaded(deviceStatus));
  }

  void _updateLocalValue(String code, dynamic value, Emitter<WaterHeaterState> emit) {
    switch (code) {
      case 'switch_1':
        if (value is bool) {
          deviceStatus = deviceStatus.copyWith(heaterSwitch: value);
        }
        break;
      default:
        break;
    }
  }

  dynamic _getValueByCode(String code) {
    switch (code) {
      case 'switch_1':
        return deviceStatus.heaterSwitch;

      default:
        return null;
    }
  }

  FutureOr<void> _fetchWaterHeaterStatus(
      WaterHeaterFetchStatusEvent event, Emitter<WaterHeaterState> emit) async {
    emit(WaterHeaterLoadingState());

    try {
      final status = await DevicesManagementApi().getDeviceStatus(event.deviceId);
      deviceStatus = WaterHeaterStatusModel.fromJson(event.deviceId, status.status);

      if (deviceStatus.countdownHours > 0 || deviceStatus.countdownMinutes > 0) {
        final remainingDuration = Duration(
          hours: deviceStatus.countdownHours,
          minutes: deviceStatus.countdownMinutes,
        );

        emit(WaterHeaterScheduleViewState(
          scheduleMode: deviceStatus.scheduleMode,
          hours: deviceStatus.countdownHours,
          minutes: deviceStatus.countdownMinutes,
          isActive: true,
          countdownRemaining: remainingDuration,
        ));

        // _startCountdown(remainingDuration, emit);
      } else {
        emit(WaterHeaterScheduleViewState(
          scheduleMode: deviceStatus.scheduleMode,
          hours: 0,
          minutes: 0,
          isActive: false,
        ));
      }
    } catch (e) {
      emit(WaterHeaterFailedState(error: e.toString()));
    }
  }

  void _startCountdown(Duration duration, Emitter<WaterHeaterState> emit) {
    _timer?.cancel();

    _timer = Timer.periodic(const Duration(seconds: 1), (timer) {
      final state = this.state as WaterHeaterScheduleViewState;
      final remaining = state.countdownRemaining! - const Duration(seconds: 1);

      if (remaining.isNegative || remaining == Duration.zero) {
        _timer?.cancel();
        emit(WaterHeaterScheduleViewState(
          scheduleMode: state.scheduleMode,
          hours: 0,
          minutes: 0,
          isActive: false,
          countdownRemaining: Duration.zero,
        ));
      } else {
        emit(WaterHeaterScheduleViewState(
          scheduleMode: state.scheduleMode,
          hours: remaining.inHours,
          minutes: remaining.inMinutes % 60,
          isActive: true,
          countdownRemaining: remaining,
        ));
      }
    });
  }

  FutureOr<void> _stopScheduleEvent(
    StopScheduleEvent event,
    Emitter<WaterHeaterState> emit,
  ) {
    _timer?.cancel();
    deviceStatus = deviceStatus.copyWith(
      countdownHours: 0,
      countdownMinutes: 0,
      scheduleMode: ScheduleModes.countdown,
    );
    emit(const WaterHeaterScheduleViewState(
      scheduleMode: ScheduleModes.countdown,
      hours: 0,
      minutes: 0,
      isActive: false,
    ));
  }

  @override
  Future<void> close() {
    _timer?.cancel();
    return super.close();
  }
}
